# (The MIT License)
#
# Copyright (c) 2016 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
blackhole.application.

This module houses methods and functionality to start the server.
"""

import asyncio
import logging
import signal
import os
import sys

from blackhole.config import Config, config_test, parse_cmd_args
from blackhole.control import stop_servers, start_servers, setgid, setuid
from blackhole.daemon import Daemon
from blackhole.exceptions import ConfigException, DaemonException
from blackhole.logs import configure_logs


def run():
    """Create the asyncio loop and start the server."""
    args = parse_cmd_args(sys.argv[1:])
    configure_logs(args)
    logger = logging.getLogger('blackhole')
    if args.test:
        config_test(args)
    try:
        config = Config(args.config_file).load().test()
    except ConfigException as err:
        logger.fatal(err)
        raise SystemExit(os.EX_USAGE)
    if args.background and not config.pidfile:
        logger.fatal('Cannot run in the background without a pidfile.')
        raise SystemExit(os.EX_USAGE)
    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal.SIGINT, loop.stop)
    start_servers()
    setgid()
    setuid()
    if args.background:
        try:
            Daemon(config.pidfile).daemonize()
        except DaemonException as err:
            stop_servers()
            logger.fatal(err)
            raise SystemExit(os.EX_USAGE)
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass
    stop_servers()
    raise SystemExit(os.EX_OK)



# (The MIT License)
#
# Copyright (c) 2016 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Configuration structure."""


import argparse
import getpass
import grp
import inspect
import logging
import os
import pwd
import re
import socket

from blackhole.exceptions import ConfigException


__version__ = __import__('blackhole').__version__


def parse_cmd_args(args):
    """
    Parse arguments from the command line.

    :param args:
    :type args: list
    """
    parser = argparse.ArgumentParser('blackhole')
    parser.add_argument('-v', '--version', action='version',
                        version=__version__)
    parser.add_argument('-c', '--conf', type=str,
                        default='/etc/blackhole.conf',
                        dest='config_file', metavar='/etc/blackhole.conf',
                        help='override the default configuration options')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-t', '--test', dest='test', action='store_true',
                       help='perform a configuration test')
    group.add_argument('-d', '--debug', dest='debug', action='store_true',
                       help='enable debugging mode')
    group.add_argument('-b', '--background', dest='background',
                       action='store_true',
                       help='run in the background')
    return parser.parse_args(args)


def config_test(args):
    """
    Test the validity of the configuration file content.

    .. note::

       Problems with the configuration will be written to the console using
       the `logging` module.

       Calls `sys.exit` upon an error.

    :param args: arguments parsed from `argparse`.
    :type args: `argparse.Namespace`
    """
    logger = logging.getLogger('blackhole.config_test')
    logger.setLevel(logging.INFO)
    try:
        conf = Config(args.config_file).load().test()
    except ConfigException as err:
        logger.fatal(err)
        raise SystemExit(os.EX_USAGE)
    logger.info('blackhole: %s syntax is OK', args.config_file)
    logger.info('blackhole: %s test was successful', args.config_file)
    if len(conf.tls_listen) > 0 and not conf.tls_dhparams:
        logger.warn('TLS is enabled but no Diffie Hellman phemeral parameters '
                    'file was provided.')
    raise SystemExit(os.EX_OK)


class Singleton(type):
    """A singleton for `blackhole.config.Config`."""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """A singleton for `blackhole.config.Config`."""
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args,
                                                                 **kwargs)
        return cls._instances[cls]


class Config(metaclass=Singleton):
    """
    Configuration module.

    Default values are provided as well as self-test functionality
    to sanity check configuration.
    """

    config_file = None
    _listen = []
    _tls_listen = []
    _user = None
    _group = None
    _timeout = 60
    _tls_key = None
    _tls_cert = None
    _tls_dhparams = None
    _pidfile = '/tmp/blackhole.pid'
    _delay = None
    _mode = 'accept'
    _max_message_size = 512000
    _dynamic_switch = None

    def __init__(self, config_file="/etc/blackhole.conf"):
        """
        Initialise the configuration.

        :param config_file: The configuration file,
                            default '/etc/blackhole.conf'
        :type config_file: str
        """
        self.config_file = config_file
        self.user = getpass.getuser()
        self.group = grp.getgrgid(os.getgid()).gr_name

    def load(self):
        """
        Load the configuration file and parse.

        Spaces, single and double quotes will be stripped. Lines beginning in
        # will be ignored.

        :returns: obj -- An instance of `blackhole.config.Config`
        """
        if self.config_file is None:
            return self
        if not os.access(self.config_file, os.R_OK):
            msg = 'Config file does not exist or is not readable.'
            raise ConfigException(msg)
        for line in open(self.config_file, 'r').readlines():
            line = line.strip()
            if line.startswith('#'):
                continue
            try:
                key, value = line.split('=')
            except ValueError:
                continue
            key, value = key.strip(), value.strip()
            value = value.replace('"', '').replace("'", '')
            setattr(self, key, value)
        return self

    @property
    def listen(self):
        """
        Addresses and ports to listen on.

        .. note::

            Default value is 127.0.0.1:25
        """
        return self._listen or [('127.0.0.1', 25, socket.AF_INET)]

    @listen.setter
    def listen(self, addrs):
        self._listen = self._listeners(addrs)

    def _convert_port(self, port):
        """
        Convert a port from the configuration files' string to an integer.

        :raises: ConfigException
        """
        try:
            _ = int(port)
        except ValueError:
            raise ConfigException('Cannot convert {} to an '
                                  'integer'.format(port))
        return int(port)

    def _listeners(self, addresses):
        """
        Convert listener lines from the configuration to usable values.

        :param addresses:
        :type addresses: str -- e.g. '127.0.0.1:25, 10.0.0.1:25'
        :returns: list or None
        """
        addrs = []
        _addrs = addresses.split(',')
        if len(_addrs) == 0:
            return
        for addr in _addrs:
            port = addr.split(':')[-1].strip()
            name = addr.replace(':{}'.format(port), '').strip()
            af = socket.AF_INET
            if ':' in name:
                af = socket.AF_INET6
            host = (name, self._convert_port(port), af)
            addrs.append(host)
        return addrs

    @property
    def tls_listen(self):
        """
        Addresses and ports to listen on for SSL/TLS connections
        """
        return self._tls_listen or []

    @tls_listen.setter
    def tls_listen(self, addrs):
        self._tls_listen = self._listeners(addrs)

    @property
    def user(self):
        """
        A UNIX user.

        .. note::

           Defaults to the current user.

        :returns: str -- A UNIX user.
        """
        return self._user

    @user.setter
    def user(self, user):
        self._user = user

    @property
    def group(self):
        """
        A UNIX group.

        .. note::

           Defaults to the current group.

        :returns: str -- A UNIX group.
        """
        return self._group

    @group.setter
    def group(self, group):
        self._group = group

    @property
    def timeout(self):
        """
        A timeout in seconds.

        .. note::

           Defaults to 60 seconds.
           Cannot be more 180 seconds for security (denial of service).

        :returns: int -- A timeout in seconds.
        """
        return int(self._timeout)

    @timeout.setter
    def timeout(self, timeout):
        self._timeout = timeout

    @property
    def tls_key(self):
        """
        A TLS key file.

        :returns: str
        """
        return self._tls_key

    @tls_key.setter
    def tls_key(self, tls_key):
        self._tls_key = tls_key

    @property
    def tls_cert(self):
        """
        A TLS certificate file.

        :returns: str
        """
        return self._tls_cert

    @tls_cert.setter
    def tls_cert(self, tls_cert):
        self._tls_cert = tls_cert

    @property
    def tls_dhparams(self):
        """
        Diffie Hellman ephemeral parameters.

        :returns: str
        """
        return self._tls_dhparams

    @tls_dhparams.setter
    def tls_dhparams(self, tls_dhparams):
        self._tls_dhparams = tls_dhparams

    @property
    def pidfile(self):
        """
        A path to store the pid.

        :returns: str -- A filesystem path.
        """
        return self._pidfile

    @pidfile.setter
    def pidfile(self, pidfile):
        self._pidfile = pidfile

    @property
    def delay(self):
        """
        A delay in seconds.

        .. note::

           Defaults to None.
           Cannot be higher than 60 seconds for security (denial of service).

        :returns: int or None -- A delay in seconds.
        """
        if self._delay is not None:
            return int(self._delay)
        return None

    @delay.setter
    def delay(self, delay):
        self._delay = delay

    @property
    def mode(self):
        """
        The mode with which to respond.

        .. note::

           Defaults to 'accept'.
           Options: 'accept', 'bounce' and 'random'.

        :returns: str -- The server response mode.
        """
        return self._mode

    @mode.setter
    def mode(self, mode):
        self._mode = mode.lower()

    @property
    def max_message_size(self):
        """
        The maximum size, in bytes, of a message.

        .. note::

           Default 512000 bytes (512 KB).
        :returns: int
        """
        if self._max_message_size is not None:
            return int(self._max_message_size)

    @max_message_size.setter
    def max_message_size(self, size):
        self._max_message_size = size

    @property
    def dynamic_switch(self):
        """
        Enable or disable dynamic switches.

        .. note::

           Allowed values are true and false.
           Default: true
        """
        if self._dynamic_switch is None:
            return True
        return self._dynamic_switch

    @dynamic_switch.setter
    def dynamic_switch(self, switch):
        if switch.lower() == 'false':
            self._dynamic_switch = False
        elif switch.lower() == 'true':
            self._dynamic_switch = True
        else:
            raise ConfigException(''''{}' is not valid. Options are true or '''
                                  '''false'''.format(switch))

    def test(self):
        """
        Test configuration validity.

        .. note::

           Uses the magic of `inspect.getmembers` to introspect methods
           beginning with 'test_' and calling them.

        :returns: obj -- An instance of `blackhole.config.Config`.
        """
        members = inspect.getmembers(self, predicate=inspect.ismethod)
        for name, _ in members:
            if name.startswith('test_'):
                getattr(self, name)()
        return self

    def test_same_listeners(self):
        """
        Test that multiple listeners are not configured on the same port.

        .. note::

           IPv4 and IPv6 addresses are different sockets so they can listen on
           the same port because they have different addresses.
        """
        if len(self.listen) == 0 and len(self.tls_listen) == 0:
            return
        if set(self.listen).intersection(self.tls_listen):
            raise ConfigException('Cannot have multiple listeners on the same'
                                  'host and port')

    def test_no_listeners(self):
        """Test that at least one listener is configured."""
        if not len(self.listen) > 0 and not len(self.tls_listen) > 0:
            raise ConfigException('You need to define at least one listener')

    def _min_max_port(self, port):
        """
        The minimum and maximum allowed port.

        :param port: The port.
        :type port: int
        :raises: `blackhole.config.ConfigException`

        .. note::

           On Linux the minimum is 1 and maximum is 65535.
        """
        min_port, max_port = 1, 65535
        if port < min_port:
            msg = ('Port number {} is not usable because it is less than '
                   '{} which is the lowest available port.').format(port,
                                                                    min_port)
            raise ConfigException(msg)
        if port > max_port:
            msg = ('Port number {} is not usable because it is less than '
                   '{} which is the highest available port').format(port,
                                                                    max_port)
            raise ConfigException(msg)

    def test_port(self):
        """
        Validate port number.

        :raises: `blackhole.exceptions.ConfigException`
        """
        for host, port, af in self.listen:
            self._port_permissions(host, port, af)

    def _port_permissions(self, host, port, af):
        """
        Validate that we have permission to use the port and it's not in use.

        :param port:
        :type port: int
        :raises: ConfigException
        """
        self._min_max_port(port)
        if os.getuid() is not 0 and port < 1024:
            msg = 'You do not have permission to use port {}'.format(port)
            raise ConfigException(msg)
        sock = socket.socket(af, socket.SOCK_STREAM)
        try:
            sock.bind((host, port))
        except OSError as err:
            errmsg = err.strerror.lower()
            msg = 'Could not use port {}, {}'.format(port, errmsg)
            raise ConfigException(msg)
        finally:
            sock.close()
            del sock

    def test_user(self):
        """
        Validate user exists in UNIX password database.

        :raises: `blackhole.exceptions.ConfigException`

        .. note::

           Defaults to `getpass.getuser` if no user is specified.
        """
        try:
            pwd.getpwnam(self.user)
        except KeyError:
            msg = '{} is not a valid user.'.format(self._user)
            raise ConfigException(msg)

    def test_group(self):
        """
        Validate group exists in UNIX group database.

        :raises: `blackhole.exceptions.ConfigException`

        .. note::

           Defaults to `grp.getgrgid.gr_name` if no group is specified.
        """
        try:
            grp.getgrnam(self.group)
        except KeyError:
            msg = '{} is a not a valid group.'.format(self._group)
            raise ConfigException(msg)

    def test_timeout(self):
        """
        Validate timeout - only allow a valid integer value in seconds.

        :raises: `blackhole.exceptions.ConfigException`
        """
        try:
            _ = self.timeout
        except ValueError:
            msg = '{} is not a valid number of seconds.'.format(self._timeout)
            raise ConfigException(msg)
        if self.timeout and self.timeout > 180:
            raise ConfigException('Timeout must be 180 seconds or less for '
                                  'security (denial of service).')

    def test_tls_port(self):
        """
        Validate TLS port number.

        :raises: `blackhole.exceptions.ConfigException`
        """
        if len(self.tls_listen) == 0:
            return
        for host, port, af in self.tls_listen:
            self._port_permissions(host, port, af)

    def test_tls_settings(self):
        """
        Validate TLS configuration.

        :raises: `blackhole.exceptions.ConfigException`

        .. note::

           Verifies if you provide all TLS settings, not just some.
        """
        port = True
        if not len(self.tls_listen) > 0:
            port = False
        cert = os.access(self.tls_cert, os.R_OK) if self.tls_cert else False
        key = os.access(self.tls_key, os.R_OK) if self.tls_key else False
        if (port, cert, key) == (False, False, False):
            return
        if not all((port, cert, key)):
            msg = ('To use TLS you must supply a port, certificate file '
                   'and key file.')
            raise ConfigException(msg)

    def test_tls_dhparams(self):
        """
        Validate Diffie Hellman ephemeral parameters.

        :raises: `blackhole.exceptions.ConfigException`

        .. note::

           Verifies Diffie Hellman ephemeral parameters are readable.
        """
        if self.tls_dhparams and not os.access(self.tls_dhparams, os.R_OK):
            msg = ('To use Diffie Hellman ephemeral params you must supply a '
                   'valid dhparams file.')
            raise ConfigException(msg)

    def test_delay(self):
        """
        Validate the delay period.

        Delay must be lower than the timeout.

        :raises: `blackhole.config.ConfigException`
        """
        if self.delay and self.delay >= self.timeout:
            raise ConfigException('Delay must be lower than timeout.')
        if self.delay and self.delay > 60:
            raise ConfigException('Delay must be 60 seconds or less for '
                                  'security (denial of service).')

    def test_mode(self):
        """
        Validate the response mode.

        Valid options are: 'accept', 'bounce' and 'random'.

        :raises: `blackhole.config.ConfigException`
        """
        if self.mode not in ('accept', 'bounce', 'random'):
            raise ConfigException('Mode must be accept, bounce or random.')

    def test_max_message_size(self):
        """
        Validate max_message size is an integer.

        :raises: `blackhole.exceptions.ConfigException`
        """
        try:
            _ = self.max_message_size
        except ValueError:
            size = self._max_message_size
            msg = '{} is not a valid number of bytes.'.format(size)
            raise ConfigException(msg)

    def test_pidfile(self):
        """
        Validate that the pidfile can be written to.

        :raises: `blackhole.exceptions.ConfigException`
        """
        if not self.pidfile:
            return
        try:
            open(self.pidfile, 'w+')
        except PermissionError:
            msg = ('You do not have permission to write to the pidfile.')
            raise ConfigException(msg)
        except FileNotFoundError:
            msg = ('The path to the pidfile does not exist.')
            raise ConfigException(msg)

    def test_dynamic_switch(self):
        """
        Validate that the dynamic_switch value is correct.

        :raises: `blackhole.exceptions.ConfigException`
        """
        if self._dynamic_switch is None:
            return
        if self._dynamic_switch not in ('true', 'false'):
            raise ConfigException('Allowed dynamic_switch values are true and '
                                  'false.')



# (The MIT License)
#
# Copyright (c) 2016 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
blackhole.control.

Command and control functionality for blackhole.
"""

import asyncio
import functools
import getpass
import grp
import logging
import os
import pwd
import socket
try:
    import ssl
except ImportError:
    ssl = None
from blackhole.config import Config
from blackhole.daemon import Daemon
from blackhole.smtp import Smtp


logger = logging.getLogger('blackhole.control')
_servers = []
ciphers = ['ECDHE-ECDSA-AES256-GCM-SHA384', 'ECDHE-RSA-AES256-GCM-SHA384',
           'ECDHE-ECDSA-CHACHA20-POLY1305', 'ECDHE-RSA-CHACHA20-POLY1305',
           'CDHE-ECDSA-AES128-GCM-SHA256', 'ECDHE-RSA-AES128-GCM-SHA256:',
           'ECDHE-ECDSA-AES256-SHA384', 'ECDHE-RSA-AES256-SHA384'
           'ECDHE-ECDSA-AES128-SHA256', 'ECDHE-RSA-AES128-SHA256']


def create_server(host, port, af, use_tls=False):
    """
    Create an instance of `socket.socket`, bind it and attach it to loop.

    .. note::

       Calls `sys.exit` when there is an error binding to the socket.
       If `use_tls` is passed, the SSL/TLS context will be created with
       `ssl.OP_NO_SSLv2` and `ssl.OP_NO_SSLv3`.

    :param use_tls: default False.
    :type use_tls: bool
    """
    logger = logging.getLogger('blackhole')
    config = Config()
    if use_tls:
        logger.debug('Creating server (%s, %s, TLS=True)', host, port)
    else:
        logger.debug('Creating server (%s, %s)', host, port)
    loop = asyncio.get_event_loop()
    factory = functools.partial(Smtp)
    sock = socket.socket(af, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    if af == socket.AF_INET6:
        sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 1)
    try:
        sock.bind((host, port))
    except OSError:
        logger.fatal("Cannot bind to port %s.", port)
        raise SystemExit(os.EX_NOPERM)
    if use_tls:
        ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ctx.load_cert_chain(config.tls_cert, config.tls_key)
        ctx.options |= ssl.OP_NO_SSLv2
        ctx.options |= ssl.OP_NO_SSLv3
        ctx.set_ciphers(':'.join(ciphers))
        if config.tls_dhparams:
            ctx.load_dh_params(config.tls_dhparams)
        else:
            logger.warn('TLS is enabled but no Diffie Hellman phemeral '
                        'parameters file was provided.')
    else:
        ctx = None
    server = loop.create_server(factory, sock=sock, ssl=ctx)
    _servers.append(loop.run_until_complete(server))


def start_servers():
    """Create each server listener and bind to the socket."""
    config = Config()
    logger.debug('Starting...')
    for host, port, af in config.listen:
        create_server(host, port, af)
    if len(config.tls_listen) > 0 and config.tls_cert and config.tls_key:
        for host, port, af in config.tls_listen:
            create_server(host, port, af, use_tls=True)


def stop_servers():
    """
    Stop the listeners.

    :raises: SystemExit
    """
    loop = asyncio.get_event_loop()
    conf = Config()
    logger.debug('Stopping...')
    for _ in range(len(_servers)):
        server = _servers.pop()
        server.close()
        loop.run_until_complete(server.wait_closed())
    for task in asyncio.Task.all_tasks():
        task.cancel()
    loop.close()
    daemon = Daemon(conf.pidfile)
    if daemon.pid:
        del daemon.pid


def setgid():
    """
    Change group.

    Drop from root privileges down to a less privileged group.

    .. note::

       MUST be called BEFORE setuid, not after.
    """
    config = Config()
    if config.group == grp.getgrgid(os.getgid()).gr_name:
        logger.debug('Group in config is the same as current group, skipping.')
        return
    try:
        os.setgid(grp.getgrnam(config.group).gr_gid)
    except KeyError:
        logger.error("Group '%s' does not exist", config.group)
        raise SystemExit(os.EX_USAGE)
    except PermissionError:
        logger.error("You do not have permission to switch to group '%s'",
                     config.group)
        raise SystemExit(os.EX_NOPERM)


def setuid():
    """
    Change user.

    Drop from root privileges down to a less privileged user.

    .. note::

       MUST be called AFTER setgid, not before.
    """
    config = Config()
    if config.user == getpass.getuser():
        logger.debug('User in config is the same as current user, skipping.')
        return
    try:
        os.setuid(pwd.getpwnam(config.user).pw_uid)
    except KeyError:
        logger.error("User '%s' does not exist", config.user)
        raise SystemExit(os.EX_USAGE)
    except PermissionError:
        logger.error("You do not have permission to switch to user '%s'",
                     config.user)
        raise SystemExit(os.EX_NOPERM)



# (The MIT License)
#
# Copyright (c) 2016 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
blackhole.smtp.

This module contains the Smtp protocol.
"""


import asyncio
import base64
from email.utils import make_msgid as message_id
import inspect
import logging
import random

from blackhole.config import Config
from blackhole.utils import mailname


logger = logging.getLogger('blackhole.smtp')


class Smtp(asyncio.StreamReaderProtocol):
    """The class responsible for handling SMTP/SMTPS commands."""

    bounce_responses = {
        450: 'Requested mail action not taken: mailbox unavailable',
        451: 'Requested action aborted: local error in processing',
        452: 'Requested action not taken: insufficient system storage',
        458: 'Unable to queue message',
        521: 'Machine does not accept mail',
        550: 'Requested action not taken: mailbox unavailable',
        551: 'User not local',
        552: 'Requested mail action aborted: exceeded storage allocation',
        553: 'Requested action not taken: mailbox name not allowed',
        571: 'Blocked',
    }
    """A dictionary of response codes and messages for bouncing mail."""

    _delay = None
    _max_delay = 60
    _mode = None

    def __init__(self):
        """
        Initialise the SMTP parotocol.

        .. note::

           Loads the configuration, defines the server's FQDN and generates
           an RFC 2822 Message-ID.
        """
        self.loop = asyncio.get_event_loop()
        super().__init__(
            asyncio.StreamReader(loop=self.loop),
            client_connected_cb=self._client_connected_cb,
            loop=self.loop)
        self.config = Config()
        self.fqdn = mailname()
        self.message_id = message_id()

    def connection_made(self, transport):
        """
        Tie a connection to blackhole to the SMTP protocol.

        :param transport:
        :type transport: `asyncio.transport.Transport`
        """
        super().connection_made(transport)
        self.peer = transport.get_extra_info('peername')
        logger.debug('Peer %s connected', repr(self.peer))
        self.transport = transport
        self.connection_closed = False
        self._handler_coroutine = self.loop.create_task(self._handle_client())

    def _client_connected_cb(self, reader, writer):
        """
        Callback that binds a stream reader and writer to the SMTP Protocol.

        :param reader:
        :type reader: `asyncio.streams.StreamReader`
        :param writer:
        :type writer: `asyncio.streams.StreamWriter`
        """
        self._reader = reader
        self._writer = writer

    def connection_lost(self, exc):
        """Callback for when a connection is closed or lost."""
        logger.debug('Peer %s disconnected', repr(self.peer))
        super().connection_lost(exc)
        self._connection_closed = True

    async def _handle_client(self):
        """
        Handle a connection to the server.

        This method greets the client and then accepts and handles each line
        the client sends, passing off to the currect verb handler.

        Client timeout is also managed and handled here.
        """
        await self.greet()
        while not self.connection_closed:
            line = await self.wait()
            logger.debug('RECV %s', line)
            line = line.decode('utf-8').rstrip('\r\n')
            handler = self.lookup_handler(line)
            if handler:
                await handler()
            else:
                await self.push(502, '5.5.2 Command not recognised')

    def get_auth_members(self):
        """
        Get a list of available AUTH mechanisms.

        :returns: list -- AUTH mechanisms.
        """
        members = inspect.getmembers(self, predicate=inspect.ismethod)
        cmds = []
        for cmd, _ in members:
            if cmd.startswith('auth_'):
                cmd = cmd.replace('auth_', '').replace('_', '-')
                cmds.append(cmd)
        return cmds

    def lookup_auth_handler(self, line):
        """
        Look up a handler for the received AUTH mechanism.

        :param line:
        :type line: str
        :returns: `blackhole.smtp.Smtp.auth_MECHANISM`.
        """
        parts = line.split(' ')
        if len(parts) < 2:
            return None
        mechanism = parts[1].upper()
        if mechanism == 'CRAM-MD5':
            return self.auth_CRAM_MD5
        if mechanism not in self.get_auth_members():
            return self.do_UNKNOWN
        if len(parts) == 3 and mechanism == 'PLAIN':
            return self._auth_success
        return getattr(self, 'auth_{}'.format(mechanism.upper()),
                       self.do_UNKNOWN)

    async def help_AUTH(self):
        """Send help for AUTH mechanisms."""
        mechanisms = ' '.join(self.get_auth_members())
        await self.push(250, 'Syntax: AUTH {}'.format(mechanisms))

    async def auth_LOGIN(self):
        """Handle an AUTH LOGIN request."""
        await self.push(334, 'VXNlcm5hbWU6')
        line = await self.wait()
        logger.debug('RECV %s', line)
        await self._auth_success()

    async def auth_CRAM_MD5(self):
        """Handle an AUTH CRAM-MD5 request."""
        message_id = base64.b64encode(self.message_id.encode('utf-8'), b'==')
        await self.push(334, message_id)
        line = await self.wait()
        logger.debug('RECV %s', line)
        await self._auth_success()

    async def wait(self):
        """
        Wait for data from the client.

        .. note::

           Also handles client timeouts if they wait too long before sending
           data.

        :returns: str
        """
        while not self.connection_closed:
            try:
                line = await asyncio.wait_for(self._reader.readline(),
                                              self.config.timeout,
                                              loop=self.loop)
            except asyncio.TimeoutError:
                await self.timeout()
            return line

    async def auth_PLAIN(self):
        """Handle an AUTH PLAIN request."""
        await self.push(334, ' ')
        line = await self.wait()
        logger.debug('RECV %s', line)
        await self._auth_success()

    async def _auth_success(self):
        """Send an authentication successful response."""
        await self.push(235, '2.7.0 Authentication successful')

    async def timeout(self):
        """
        Timeout a client connection.

        Sends the 421 timeout message to the client and closes the connection.
        """
        logger.debug('%s timed out, no data received for %d seconds',
                     repr(self.peer), self.config.timeout)
        await self.push(421, 'Timeout')
        await self.close()

    async def close(self):
        """Close the connection from the client."""
        logger.debug('Closing connection: %s', repr(self.peer))
        if self._writer:
            self._writer.close()
        self._connection_closed = True

    def lookup_handler(self, line):
        """
        Look up the SMTP VERB against a handler.

        :param line:
        :type line: str -- e.g. HELO blackhole.io
        :returns: `blackhole.smtp..Smtp.do_VERB`,
                  `blackhole.smtp.Smtp.auth_MECHANISM` or
                  `blackhole.smtp..Smtp.help_VERB`.
        """
        parts = line.split(None, 1)
        if parts:
            if parts[0].lower() == 'help':
                return self.lookup_help_handler(parts)
            if parts[0].lower() == 'auth':
                return self.lookup_auth_handler(line)
            else:
                return self.lookup_verb_handler(parts[0])
        return self.do_UNKNOWN

    def lookup_help_handler(self, parts):
        """
        Look up a help handler for the SMTP VERB.

        :param parts:
        :type parts: list
        :returns: `blackhole.smtp.Smtp.help_VERB`.
        """
        if len(parts) > 1:
            cmd = 'help_{}'.format(parts[1].upper())
        else:
            cmd = 'do_HELP'
        return getattr(self, cmd, self.help_UNKNOWN)

    def lookup_verb_handler(self, verb):
        """
        Look up a handler for the SMTP VERB.

        :param verb:
        :type verb: str
        :returns: `blackhole.smtp.Smtp.do_VERB`.
        """
        return getattr(self, 'do_{}'.format(verb.upper()), self.do_UNKNOWN)

    async def push(self, code, msg):
        """
        Write a response code and message to the client.

        :param code:
        :type int: SMTP code, i.e. 250.
        :param msg:
        :type msg: The message for the SMTP code.
        """
        response = "{} {}\r\n".format(code, msg).encode('utf-8')
        logger.debug('SEND %s', response)
        self._writer.write(response)
        await self._writer.drain()

    async def greet(self):
        """Send a greeting to the client."""
        await self.push(220, '{} ESMTP'.format(self.fqdn))

    def get_help_members(self):
        """
        Get a list of HELP handlers for verbs.

        :returns: list -- help handler names.
        """
        members = inspect.getmembers(self, predicate=inspect.ismethod)
        cmds = []
        for cmd, _ in members:
            if cmd.startswith('help_') and not cmd == 'help_UNKNOWN':
                cmds.append(cmd.replace('help_', ''))
        return cmds

    async def do_HELP(self):
        """Send a response to the HELP verb."""
        msg = ' '.join(self.get_help_members())
        await self.push(250, 'Supported commands: {}'.format(msg))

    async def help_HELO(self):
        """Send help for HELO verb."""
        await self.push(250, 'Syntax: HELO domain.tld')

    async def do_HELO(self):
        """Send response to HELO verb."""
        await self.push(250, 'OK')

    async def help_EHLO(self):
        """Send help for the EHLO verb."""
        await self.push(250, 'Syntax: EHLO domain.tld')

    async def do_EHLO(self):
        """Send response to EHLO verb."""
        response = "250-{}\r\n".format(self.fqdn).encode('utf-8')
        self._writer.write(response)
        logger.debug('SENT %s', response)
        auth = ' '.join(self.get_auth_members())
        responses = ('250-HELP', '250-PIPELINING', '250-AUTH {}'.format(auth),
                     '250-SIZE {}'.format(self.config.max_message_size),
                     '250-VRFY', '250-ETRN', '250-ENHANCEDSTATUSCODES',
                     '250-8BITMIME', '250-SMTPUTF8', '250 DSN', )
        for response in responses:
            response = "{}\r\n".format(response).encode('utf-8')
            logger.debug("SENT %s", response)
            self._writer.write(response)
        await self._writer.drain()

    async def help_MAIL(self):
        """Send help for the MAIL TO verb."""
        await self.push(250, 'Syntax: MAIL FROM: <address>')

    async def do_MAIL(self):
        """Send response to MAIL TO verb."""
        await self.push(250, '2.1.0 OK')

    async def help_RCPT(self):
        """Send response to the RCPT TO verb."""
        await self.push(250, 'Syntax: RCPT TO: <address>')

    async def do_RCPT(self):
        """Send response to RCPT TO verb."""
        await self.push(250, '2.1.5 OK')

    async def help_DATA(self):
        """Send help for the DATA verb."""
        await self.push(250, 'Syntax: DATA')

    def process_header(self, line):
        """
        Process dynamic switch email headers.

        Reads x-blackhole-delay and x-blackhole-mode headers and re-configures
        on-the-fly how the email is handled based on these headers.

        :param line:
        :type line: str -- an email header
        """
        logger.debug('HEADER RECV: %s', line)
        if self.config.dynamic_switch is False:
            logger.debug('Dynamic switches disabled, ignoring')
            return
        key, value = line.split(':')
        key, value = key.lower().strip(), value.lower().strip()
        if key == 'x-blackhole-delay':
            self.delay = value
        if key == 'x-blackhole-mode':
            self.mode = value

    async def response_from_mode(self):
        """
        Send a response based on the configured response mode.

        Response mode is configured in configuration file and can be overridden
        by email headers, if enabled.
        """
        logger.debug('MODE: %s', self.mode)
        if self.mode == 'bounce':
            key = random.choice(list(self.bounce_responses.keys()))
            await self.push(key, self.bounce_responses[key])
        elif self.mode == 'random':
            resps = {250: '2.0.0 OK: queued as {}'.format(self.message_id), }
            resps.update(self.bounce_responses)
            key = random.choice(list(resps.keys()))
            await self.push(key, resps[key])
        else:
            msg = '2.0.0 OK: queued as {}'.format(self.message_id)
            await self.push(250, msg)

    async def do_DATA(self):
        r"""
        Send response to DATA verb and wait for mail data.

        This method will also implement timeout management and handling after
        receiving the DATA command and no new data is received.

        This method also implements they delay functionality, delaying a
        response after the final '\r\n.\r\n' line.

        This method is also responsible handling the mode with which to
        respond to the client.
        """
        await self.push(354, 'End data with <CR><LF>.<CR><LF>')
        on_body = False
        msg = []
        while not self.connection_closed:
            line = await self.wait()
            logger.debug('RECV %s', line)
            msg.append(line)
            if line.lower().startswith(b'x-blackhole') and on_body is False:
                self.process_header(line.decode('utf-8').rstrip('\n'))
            if len(b''.join(msg)) > self.config.max_message_size:
                await self.push(552, 'Message size exceeds fixed maximum '
                                'message size')
                return
            if line == b'\n':
                on_body = True
            if line == b'.\r\n':
                break
        if self.delay:
            logger.debug('DELAYING RESPONSE: %s seconds', self.delay)
            await asyncio.sleep(self.delay)
        await self.response_from_mode()

    async def do_STARTTLS(self):
        """STARTTLS is not implemented."""
        # It's currently not possible to implement STARTTLS due to lack of
        # support in asyncio. - https://bugs.python.org/review/23749/
        await self.do_NOT_IMPLEMENTED()

    async def help_NOOP(self):
        """Send help for the NOOP verb."""
        await self.push(250, 'Syntax: NOOP')

    async def do_NOOP(self):
        """Send response to the NOOP verb."""
        await self.push(250, '2.0.0 OK')

    async def help_RSET(self):
        """Send help for the RSET verb."""
        await self.push(250, 'Syntax: RSET')

    async def do_RSET(self):
        """
        Send response to the RSET verb.

        A new message id is generated and assigned.
        """
        old_msg_id = self.message_id
        self.message_id = message_id()
        logger.debug('%s is now %s', old_msg_id, self.message_id)
        await self.push(250, '2.0.0 OK')

    async def help_VRFY(self):
        """Send help for the VRFY verb."""
        await self.push(250, 'Syntax: VRFY <address>')

    async def do_VRFY(self):
        """Send response to the VRFY verb."""
        await self.push(252, '2.0.0 OK')

    async def help_ETRN(self):
        """Send help for the ETRN verb."""
        await self.push(250, 'Syntax: ETRN')

    async def do_ETRN(self):
        """Send response to the ETRN verb."""
        await self.push(250, 'Queueing started')

    async def help_QUIT(self):
        """Send help for the QUIT verb."""
        await self.push(250, 'Syntax: QUIT')

    async def do_QUIT(self):
        """
        Send response to the QUIT verb.

        Closes the client connection.
        """
        await self.push(221, '2.0.0 Goodbye')
        self._handler_coroutine.cancel()
        await self.close()

    async def do_NOT_IMPLEMENTED(self):
        """Send a not implemented response."""
        await self.push(500, 'Not implemented')

    async def help_UNKNOWN(self):
        """Send available help verbs when an invalid verb is received."""
        msg = ' '.join(self.get_help_members())
        await self.push(501, 'Supported commands: {}'.format(msg))

    async def do_UNKNOWN(self):
        """Send response to unknown verb."""
        await self.push(502, '5.5.2 Command not recognised')

    @property
    def delay(self):
        """
        Delay after the DATA command completes.

        Value is in seconds, with a maximum value of 60 seconds.
        """
        if self._delay is not None:
            return self._delay
        if self.config.delay is not None:
            return self.config.delay
        return None

    @delay.setter
    def delay(self, values):
        logger.debug('DELAY: Dymanic delay enabled')
        value = values.split(',')
        if len(value) == 2:
            self._delay_range(value)
        elif len(value) == 1:
            self._delay_single(value[0])
        else:
            logger.debug('DELAY: Invalid value(s): %s. Skipping', values)
            return

    def _delay_range(self, value):
        """
        Generate a delay from a range provided in the email header.

        :param value:
        :type value: str  -- a list of minimum and maximum values as a string.
                     i.e. (10, 20).

        .. note::

           Converted from a string of a list to a list of integers.
        """
        min_delay, max_delay = value
        min_delay, max_delay = min_delay.strip(), max_delay.strip()
        try:
            min_delay = int(min_delay)
            max_delay = int(max_delay)
        except ValueError:
            logger.debug('DELAY: Unable to convert %s, %s to integers. '
                         'Skipping', min_delay, max_delay)
            self._delay = None
            return
        if min_delay < 0 or max_delay < 0:
            logger.debug('DELAY: A value is less than 0: %s, %s. Skipping',
                         min_delay, max_delay)
            self._delay = None
            return
        if min_delay > max_delay:
            logger.debug('Min cannot be greater than max')
            self._delay = None
            return
        if max_delay > self._max_delay:
            logger.debug('DELAY: %s is higher than %s. %s is the hard coded '
                         'maximum delay for security.', max_delay,
                         self._max_delay, self._max_delay)
            max_delay = self._max_delay
        self._delay = random.randint(min_delay, max_delay)
        logger.debug('DELAY: Set to %s from range %s-%s', self._delay,
                     min_delay, max_delay)
        return

    def _delay_single(self, value):
        """
        Generate a delay from a value provided in an email header.

        :param value:
        :type value: str -- 10

        .. note:

           Converted from a string to an integer.
        """
        try:
            value = int(value)
        except ValueError:
            logger.debug('DELAY: Unable to convert %s to an integer. Skipping',
                         value)
            self._delay = None
            return
        logger.debug(value)
        if value < 0:
            logger.debug('DELAY: %s is less than 0. Skipping', value)
            self._delay = None
            return
        if value > self._max_delay:
            logger.debug('DELAY: %s is higher than %s. %s is the hard coded '
                         'maximum delay for security.', value, self._max_delay,
                         self._max_delay)
            self._delay = self._max_delay
            return
        logger.debug('DELAY: Set to %s', value)
        self._delay = value

    @property
    def mode(self):
        """
        How to respond to an email, based on configuration.

        Reponse is configured in the configuration file or configured from
        email headers, if configured to allow that option.
        """
        if self._mode is not None:
            return self._mode
        return self.config.mode

    @mode.setter
    def mode(self, value):
        if value not in ['accept', 'bounce', 'random']:
            logger.debug('MODE: %s is an invalid. Allowed modes: (accept, '
                         'bounce, random)', value)
            self._mode = None
            return
        logger.debug('MODE: Dynamic mode enabled. Mode set to %s', value)
        self._mode = value



# (The MIT License)
#
# Copyright (c) 2016 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Blackhole is an email MTA that pipes all mail to /dev/null.

Blackhole is built on top of asyncio and utilises `async` and `await`
statements on available in Python 3.5 and above.

While Blackhole is an MTA (mail transport agent), none of the actions
performed of SMTP or SMTPS are actually processed and no email or sent or
delivered.
"""

from blackhole.application import run


if __name__ == '__main__':  # pragma: no cover
    run()



# (The MIT License)
#
# Copyright (c) 2016 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
blackhole.daemon.

Provides daemon functional to blackhole.
"""


import os
import atexit

from blackhole.exceptions import DaemonException


class Singleton(type):
    """A singleton for `blackhole.daemon.Daemon`."""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        A singleton for `blackhole.daemon.Daemon`.

        :param cls:
        type cls: str
        """
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args,
                                                                 **kwargs)
        return cls._instances[cls]


class Daemon(metaclass=Singleton):

    def __init__(self, pidfile):
        """
        Create an instance of `blackhole.daemon.Daemon`.

        :param pidfile:
        :type pidfile: str
        """
        self.pidfile = pidfile

    def daemonize(self):
        """
        Daemonize the process, using the UNIX double fork method.

        .. note::

           Registers an `atexit.register` signal to delete the pid on exit.
        """
        self.fork()
        os.chdir(os.path.sep)
        os.setsid()
        os.umask(0)
        self.fork()
        atexit.register(self._delpid)
        self.pid = os.getpid()

    def _delpid(self):
        """
        A wrapper around `del self.pid` for use with `atexit.register`.

        This is not a public method for API use.
        """
        del self.pid

    def fork(self):
        try:
            pid = os.fork()
            if pid > 0:
                raise SystemExit(os.EX_OK)
        except OSError as err:
            raise DaemonException(err.strerror)

    @property
    def pid(self):
        """
        The pid of the process, if it's been daemonised.

        .. note::

           The pid is retrieved from the filestem.
           If the pid does not exist in /proc, the pid is deleted.

        :raises: IOError if pid cannot be read from the filesystem.
        :returns: int or None if no pid.
        """
        if os.path.exists(self.pidfile):
            try:
                pid = open(self.pidfile, 'r').read().strip()
                if pid != '':
                    return int(pid)
            except (FileNotFoundError, IOError, PermissionError,
                    OSError) as err:
                raise DaemonException(err.strerror)
        return None

    @pid.setter
    def pid(self, pid):
        """
        Write the pid to the filesystem.

        :raises: `daemon.daemon.DaemonException` if writing to filesystem
                 fails.
        :param pid: the process's pid.
        :type pid: int
        """
        pid = str(pid)
        try:
            with open(self.pidfile, 'w+') as pidfile:
                pidfile.write("{}\n".format(pid))
        except (IOError, FileNotFoundError, PermissionError) as err:
            raise DaemonException(err.strerror)

    @pid.deleter
    def pid(self):
        """Delete the pid from the filesystem."""
        if os.path.exists(self.pidfile):
            os.remove(self.pidfile)



# (The MIT License)
#
# Copyright (c) 2016 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
blackhole.logs.

Configure logging for blackhole.
"""


import logging
from logging.config import dictConfig


DEBUG_FORMAT = ('[%(asctime)s] [%(levelname)s] blackhole.%(module)s:'
                '%(message)s')

LOG_CONFIG = {
    'version': 1,
    'formatters': {
        'console': {'format': '%(message)s'},
        'debug': {'format': DEBUG_FORMAT},
    },
    'handlers': {
    },
    'loggers': {
        'blackhole': {'handlers': [], 'level': logging.INFO},
    }
}

DEBUG_HANDLER = {'class': 'logging.StreamHandler',
                 'formatter': 'debug',
                 'level': logging.DEBUG}

DEFAULT_HANDLER = {'class': 'logging.StreamHandler',
                   'formatter': 'console',
                   'level': logging.INFO}


def configure_logs(args):
    """
    Configure the logging module.

    :param args:
    :type args: Parameters parsed from `argparse`.
    """
    logger_handlers = LOG_CONFIG['loggers']['blackhole']['handlers']
    if args.debug and not args.test:
        LOG_CONFIG['loggers']['blackhole']['level'] = logging.DEBUG
        LOG_CONFIG['handlers']['default_handler'] = DEBUG_HANDLER
        logger_handlers.append('default_handler')
    else:
        LOG_CONFIG['loggers']['blackhole']['level'] = logging.INFO
        LOG_CONFIG['handlers']['default_handler'] = DEFAULT_HANDLER
        logger_handlers.append('default_handler')
    dictConfig(LOG_CONFIG)



# (The MIT License)
#
# Copyright (c) 2016 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
blackhole.utils.

Provides utility functions to blackhole.
"""


import os
import socket


def mailname():
    """
    A fully qualified domain name for HELO and EHLO.

    .. note::

       Prefers content of /etc/mailname, falls back on `socket.getfqdn`.

    :returns: str -- fully qualified domain name.
    """
    mailname_file = '/etc/mailname'
    if os.path.exists(mailname_file):
        mailname_content = open(mailname_file, 'r').read().strip()
        if mailname_content != '':
            return mailname_content
    return socket.getfqdn()



import sys

from setup_helpers import require_python, get_version, long_description
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


require_python(50659568)
__version__ = get_version('blackhole/__init__.py')


class PyTest(TestCommand):

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = [
            '--doctest-modules', '--verbose',
            './blackhole', './tests'
        ]
        self.test_suite = True

    def run_tests(self):
        import pytest
        sys.exit(pytest.main(self.test_args))


entry_points = {
    'console_scripts': [
        'blackhole = blackhole.application:run',
    ]
}

tests_require = [
    'pytest',
    'pytest-asyncio'
]


setup(name='blackhole',
      version=__version__,
      url='https://blackhole.io/',
      author='Kura',
      author_email='kura@kura.io',
      maintainer='Kura',
      maintainer_email='kura@kura.io',
      description=('Blackhole is an MTA (message transfer agent) that '
                   '(figuratively) pipes all mail to /dev/null.'),
      long_description=long_description('README.rst'),
      keywords='blackhole, mta, email',
      license=long_description('LICENSE'),
      platforms=['linux'],
      packages=find_packages(exclude=["*.tests"]),
      install_requires=[],
      tests_require=tests_require,
      cmdclass={'test': PyTest},
      entry_points=entry_points,
      classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'Intended Audience :: Information Technology',
            'Intended Audience :: System Administrators',
            'License :: OSI Approved :: MIT License',
            'Operating System :: POSIX :: Linux',
            'Operating System :: Unix',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3 :: Only',
            'Topic :: Communications :: Email',
            'Topic :: Communications :: Email :: Mail Transport Agents',
            'Topic :: Education :: Testing',
            'Topic :: Internet',
            'Topic :: Software Development',
            'Topic :: Software Development :: Testing',
            'Topic :: Software Development :: Testing :: Traffic Generation',
            'Topic :: System :: Networking',
            'Topic :: System :: Systems Administration',
            'Topic :: Utilities', ],
      )
